/*
 *  Copyright 2009 Richard Nichols.
 * 
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 * 
 *       http://www.apache.org/licenses/LICENSE-2.0
 * 
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *  under the License.
 */
package com.visural.common;

import com.visural.common.apacherepack.FastDateFormat;
import java.text.ParsePosition;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.TimeZone;

/**
 * Date utilities
 * 
 * @version $Id: DateUtil.java 96 2010-10-13 09:45:27Z tibes80@gmail.com $
 * @author Richard Nichols
 */
public class DateUtil {

    /**
     * Makes a copy of the given Date. nulls are passed through.
     * @param date
     * @return
     */
    public static Date copy(Date date) {
        return (date == null ? null : new Date(date.getTime()));
    }

    /**
     * Formats a date according to the SimpleDateFormat format string provided
     * @param date the date to display
     * @param sFormat the format string according to the SimpleDateFormat class specs
     * @return the resulting formatted string
     */
    public static String format(Date date, String format) {
        return date == null ? null : FastDateFormat.getInstance(format).format(date);
    }

    public static String format(Date d, String format, TimeZone timeZone) {
        return d == null ? null : FastDateFormat.getInstance(format, timeZone).format(d);
    }

    /**
     * Returns a date in the format used by the XML standard (ISO format)
     * 
     * @param date
     * @return 
     */
    public static String formatXmlIsoDate(Date date) {
        if (date == null) {
            return null;
        }
        Calendar calendar = new GregorianCalendar(TimeZone.getTimeZone("UTC"));
        calendar.setTime(date);
        StringBuilder buffer = new StringBuilder();
        buffer.append(calendar.get(Calendar.YEAR));
        buffer.append("-");
        buffer.append(twoDigit(calendar.get(Calendar.MONTH) + 1));
        buffer.append("-");
        buffer.append(twoDigit(calendar.get(Calendar.DAY_OF_MONTH)));
        buffer.append("T");
        buffer.append(twoDigit(calendar.get(Calendar.HOUR_OF_DAY)));
        buffer.append(":");
        buffer.append(twoDigit(calendar.get(Calendar.MINUTE)));
        buffer.append(":");
        buffer.append(twoDigit(calendar.get(Calendar.SECOND)));
        buffer.append(".");
        buffer.append(twoDigit(calendar.get(Calendar.MILLISECOND) / 10));
        buffer.append("Z");
        return buffer.toString();
    }

    private static String twoDigit(int i) {
        if (i >= 0 && i < 10) {
            return "0" + String.valueOf(i);
        }
        return String.valueOf(i);
    }
    
    /**
     * Pattern for RFC1123 (http date header)
     */
    public static final String PATTERN_RFC1123 = "EEE, dd MMM yyyy HH:mm:ss zzz";

    /**
     * Return a date formatted to RFC1123 suitable for Http header
     * @param date
     * @return
     */
    public static String formatHttpDate(Date date) {
        return format(date, PATTERN_RFC1123);
    }

    /**
     * Return a date formatted to RFC1123 suitable for Http header
     * @param millis
     * @return
     */
    public static String formatHttpDate(long millis) {
        return formatHttpDate(new Date(millis));
    }

    /**
     * Parses a string into Date using the SimpleDateFormat format string provided.
     * @param dateStr
     * @param format
     * @return
     */
    public static Date parseDate(String dateStr, String... formats) {
        return parseDate(dateStr, null, formats);
    }
    
    /**
     * Parses a string into Date using the SimpleDateFormat format string provided.
     * @param dateStr
     * @param format
     * @return
     */
    public static Date parseDate(String dateStr, TimeZone timeZone, String... formats) {
        if (dateStr == null) {
            return null;
        }
        for (String format : formats) {
            SimpleDateFormat sdf = new SimpleDateFormat(format);
            if (timeZone != null) {
                sdf.setTimeZone(timeZone);
            }
            Date result = sdf.parse(dateStr, new ParsePosition(0));
            if (result != null) {
                return result;
            }
        }
        return null;
    }

    /**
     * Converts a date to a SQL timestamp
     * @param d
     * @return
     */
    public static java.sql.Timestamp dateToTimestamp(java.util.Date d) {
        if (d == null) {
            return null;
        }
        java.sql.Timestamp ts;
        ts = new java.sql.Timestamp(d.getTime());
        return ts;
    }

    /**
     * Convers a SQL timestamp to a Date
     * @param ts
     * @return
     */
    public static java.util.Date timestampToDate(java.sql.Timestamp ts) {
        if (ts == null) {
            return null;
        }
        Date d = new Date(ts.getTime());
        return d;
    }
}
